import os.path

# Get package metadata from '__about__.py' file.
about = {}
base_dir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
with open(
    os.path.join(base_dir, 'src', 'rest_framework_reactive', '__about__.py'), 'r'
) as fh:
    exec(fh.read(), about)
if '__version__' not in about:
    raise AttributeError(
        "Package's version is not defined. Please, install the package."
    )

# -- General configuration ------------------------------------------------

# The extension modules to enable.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Django REST Framework Reactive'
version = about['__version__']
release = version
author = about['__author__']
copyright = about['__copyright__']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Warn about all references where the target cannot be found.
nitpicky = True
# Except for the following:
nitpick_ignore = [
    # This is needed to prevent warnings for container types, e.g.:
    # :type foo: tuple(bool, str).
    ('py:obj', '')
]

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = 'sphinx_rtd_theme'

# Output file base name for HTML help builder.
htmlhelp_basename = 'DjangoRESTFrameworkReactivedoc'

# Configuration for intersphinx.
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}



from django.conf import settings


def get_queryobserver_settings():
    """Query observer connection configuration."""
    defaults = {
        # Observers going over these limits will emit warnings.
        'warnings': {'max_result_length': 1000, 'max_processing_time': 1.0},
        # Observers going over these limits will be stopped.
        'errors': {'max_processing_time': 20.0},
        # Throttle evaluation (in seconds). If a new update comes earlier than
        # given rate value, the evaluation will be delayed (and batched).
        # A higher value introduces more latency.
        'throttle_rate': 2,
    }
    defaults.update(getattr(settings, 'DJANGO_REST_FRAMEWORK_REACTIVE', {}))
    return defaults



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-06-07 02:42
from __future__ import unicode_literals

from django.db import migrations

# NOTE: The name of the constrait is generated by Django ORM.
UNIQUE_INDEX_NAME = 'rest_framework_reactive_item_observer_id_order_9b8adde6_uniq'


class Migration(migrations.Migration):

    dependencies = [('rest_framework_reactive', '0001_initial')]

    operations = [
        migrations.RunSQL(
            [
                # Drop old constraint.
                "ALTER TABLE rest_framework_reactive_item DROP CONSTRAINT {}".format(
                    UNIQUE_INDEX_NAME
                ),
                # Re-create the unique index.
                "CREATE UNIQUE INDEX {} ON rest_framework_reactive_item (observer_id, \"order\")".format(
                    UNIQUE_INDEX_NAME
                ),
                # Create new deferrable constraint.
                """ALTER TABLE rest_framework_reactive_item ADD CONSTRAINT
                {0} UNIQUE USING INDEX {0} DEFERRABLE INITIALLY IMMEDIATE""".format(
                    UNIQUE_INDEX_NAME
                ),
            ]
        )
    ]



import hashlib

from django.http import request as http_request

# Observable query parameter name.
OBSERVABLE_QUERY_PARAMETER = 'observe'


class Request(http_request.HttpRequest):
    """Request used by the query observer to interact with the viewsets.

    This request class is picklable.
    """

    def __init__(self, viewset_class, viewset_method, request, args=None, kwargs=None):
        """
        :param request: The original API request
        """

        super().__init__()

        self.viewset_class = viewset_class
        self.viewset_method = viewset_method
        self.args = args or []
        self.kwargs = kwargs or {}

        # Copy relevant fields from the original request.
        self.method = request.method
        self.META = {}
        for key, value in request._request.META.items():
            if isinstance(value, str):
                self.META[key] = value
        self.GET = request._request.GET.copy()
        if OBSERVABLE_QUERY_PARAMETER in self.GET:
            # Remove the original observe query parameter.
            del self.GET[OBSERVABLE_QUERY_PARAMETER]
        self.path = request._request.path
        self.path_info = request._request.path_info
        self._force_auth_user = request.user
        self._observe_id = None

    @property
    def observe_id(self):
        """Unique identifier that identifies the observer."""
        if self._observe_id is None:
            hasher = hashlib.sha256()
            hasher.update(self.viewset_class.__module__.encode('utf8'))
            hasher.update(self.viewset_class.__name__.encode('utf8'))
            hasher.update(self.viewset_method.encode('utf8'))
            # Arguments do not need to be taken into account as they are
            # derived from the request path, which is already accounted for.
            for key in sorted(self.GET.keys()):
                hasher.update(key.encode('utf8'))
                hasher.update(self.GET[key].encode('utf8'))
            hasher.update(self.path.encode('utf8'))
            hasher.update(self.path_info.encode('utf8'))
            if self._force_auth_user is not None:
                hasher.update(
                    (str(self._force_auth_user.id) or 'anonymous').encode('utf8')
                )
            else:
                hasher.update(b'anonymous')
            self._observe_id = hasher.hexdigest()

        return self._observe_id

    def __getstate__(self):
        return {
            'viewset_class': self.viewset_class,
            'viewset_method': self.viewset_method,
            'args': self.args,
            'kwargs': self.kwargs,
            'method': self.method,
            'META': self.META,
            'GET': self.GET,
            'path': self.path,
            'path_info': self.path_info,
            'user': self._force_auth_user,
            'observe_id': self._observe_id,
        }

    def __setstate__(self, state):
        self.viewset_class = state['viewset_class']
        self.viewset_method = state['viewset_method']
        self.args = state['args']
        self.kwargs = state['kwargs']
        self.method = state['method']
        self.META = state['META']
        self.GET = state['GET']
        self.path = state['path']
        self.path_info = state['path_info']
        self._force_auth_user = state['user']
        self._observe_id = state['observe_id']

    def __repr__(self):
        return '<Request: viewset={viewset} method={method} path={path} query={get}>'.format(
            viewset=repr(self.viewset_class),
            method=self.viewset_method,
            path=self.path,
            get=repr(self.GET),
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-05-12 12:00
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Dependency',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('table', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('primary_key', models.CharField(max_length=200)),
                ('order', models.IntegerField()),
                ('data', django.contrib.postgres.fields.jsonb.JSONField()),
            ],
            options={'ordering': ('observer', 'order')},
        ),
        migrations.CreateModel(
            name='Observer',
            fields=[
                (
                    'id',
                    models.CharField(max_length=64, primary_key=True, serialize=False),
                ),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('new', 'New'),
                            ('observing', 'Observing'),
                            ('stopped', 'Stopped'),
                        ],
                        default='new',
                        max_length=15,
                    ),
                ),
                ('request', models.BinaryField()),
                ('last_evaluation', models.DateTimeField(auto_now_add=True)),
                ('poll_interval', models.IntegerField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Subscriber',
            fields=[
                (
                    'session_id',
                    models.CharField(max_length=100, primary_key=True, serialize=False),
                ),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.AddField(
            model_name='observer',
            name='subscribers',
            field=models.ManyToManyField(to='rest_framework_reactive.Subscriber'),
        ),
        migrations.AddField(
            model_name='item',
            name='observer',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='items',
                to='rest_framework_reactive.Observer',
            ),
        ),
        migrations.AddField(
            model_name='dependency',
            name='observer',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='dependencies',
                to='rest_framework_reactive.Observer',
            ),
        ),
        migrations.AlterUniqueTogether(
            name='item',
            unique_together=set([('observer', 'primary_key'), ('observer', 'order')]),
        ),
        migrations.AlterUniqueTogether(
            name='dependency', unique_together=set([('observer', 'table')])
        ),
    ]



"""
The models defined here are only used during testing.
"""
from django.db import models


class ExampleItem(models.Model):
    enabled = models.BooleanField()
    name = models.CharField(max_length=30)

    class Meta:
        permissions = (("view_exampleitem", "Can view example item"),)
        default_permissions = ()
        ordering = ['pk']


class ExampleSubItem(models.Model):
    parent = models.ForeignKey(ExampleItem, on_delete=models.CASCADE)
    enabled = models.BooleanField()

    class Meta:
        permissions = (("view_examplesubitem", "Can view example sub item"),)
        default_permissions = ()
        ordering = ['pk']


class ExampleM2MItem(models.Model):
    value = models.IntegerField(default=1)
    items = models.ManyToManyField(ExampleItem, related_name='items')

    class Meta:
        ordering = ['pk']



import asyncio
import collections
import pickle

from channels.consumer import AsyncConsumer
from channels.db import database_sync_to_async
from channels.generic.websocket import JsonWebsocketConsumer
from django.core.cache import cache

from .connection import get_queryobserver_settings
from .models import Observer, Subscriber
from .observer import QueryObserver
from .protocol import *

# Maximum number of cached observer executors.
MAX_CACHED_EXECUTORS = 1024
# Throttle constants
THROTTLE_CACHE_PREFIX = 'rest_framework_reactive_throttle_'


def throttle_cache_key(observer_id):
    return '{}{}'.format(THROTTLE_CACHE_PREFIX, observer_id)


class MainConsumer(AsyncConsumer):
    """Consumer for polling observers."""

    async def observer_orm_notify(self, message):
        """Process notification from ORM."""

        @database_sync_to_async
        def get_observers(table):
            # Find all observers with dependencies on the given table.
            return list(
                Observer.objects.filter(
                    dependencies__table=table, subscribers__isnull=False
                )
                .distinct('pk')
                .values_list('pk', flat=True)
            )

        observers_ids = await get_observers(message['table'])

        for observer_id in observers_ids:
            await self.channel_layer.send(
                CHANNEL_WORKER, {'type': TYPE_EVALUATE, 'observer': observer_id}
            )

    async def observer_poll(self, message):
        """Poll observer after a delay."""
        # Sleep until we need to notify the observer.
        await asyncio.sleep(message['interval'])

        # Dispatch task to evaluate the observable.
        await self.channel_layer.send(
            CHANNEL_WORKER, {'type': TYPE_EVALUATE, 'observer': message['observer']}
        )


class WorkerConsumer(AsyncConsumer):
    """Worker consumer."""

    def __init__(self, *args, **kwargs):
        """Construct observer worker consumer."""
        self._executor_cache = collections.OrderedDict()
        super().__init__(*args, **kwargs)

    async def _evaluate(self, observer_id):
        # Get Observer from database.
        @database_sync_to_async
        def get_observer(observer_id):
            try:
                return Observer.objects.only('pk', 'request').get(pk=observer_id)
            except Observer.DoesNotExist:
                return None

        observer = await get_observer(observer_id)
        if not observer:
            return

        # Get QueryObserver executor from cache and evaluate.
        try:
            executor = self._executor_cache[observer.pk]
            self._executor_cache.move_to_end(observer.pk)
        except KeyError:
            executor = QueryObserver(pickle.loads(observer.request))
            self._executor_cache[observer.pk] = executor
            if len(self._executor_cache) > MAX_CACHED_EXECUTORS:
                self._executor_cache.popitem(last=False)

        await executor.evaluate()

    async def observer_evaluate(self, message):
        """Execute observer evaluation on the worker or throttle."""
        observer_id = message['observer']
        throttle_rate = get_queryobserver_settings()['throttle_rate']
        if throttle_rate <= 0:
            await self._evaluate(observer_id)
            return

        cache_key = throttle_cache_key(observer_id)
        try:
            count = cache.incr(cache_key)
            # Ignore if delayed observer already scheduled.
            if count == 2:
                await self.channel_layer.send(
                    CHANNEL_MAIN,
                    {
                        'type': TYPE_POLL,
                        'observer': observer_id,
                        'interval': throttle_rate,
                    },
                )
        except ValueError:
            count = cache.get_or_set(cache_key, default=1, timeout=throttle_rate)
            # Ignore if cache was set and increased in another thread.
            if count == 1:
                await self._evaluate(observer_id)


class ClientConsumer(JsonWebsocketConsumer):
    """Client consumer."""

    def websocket_connect(self, message):
        """Called when WebSocket connection is established."""
        self.session_id = self.scope['url_route']['kwargs']['subscriber_id']
        super().websocket_connect(message)

        # Create new subscriber object.
        Subscriber.objects.get_or_create(session_id=self.session_id)

    @property
    def groups(self):
        """Groups this channel should add itself to."""
        if not hasattr(self, 'session_id'):
            return []

        return [GROUP_SESSIONS.format(session_id=self.session_id)]

    def disconnect(self, code):
        """Called when WebSocket connection is closed."""
        Subscriber.objects.filter(session_id=self.session_id).delete()

    def observer_update(self, message):
        """Called when update from observer is received."""
        # Demultiplex observer update into multiple messages.
        for action in ('added', 'changed', 'removed'):
            for item in message[action]:
                self.send_json(
                    {
                        'msg': action,
                        'observer': message['observer'],
                        'primary_key': message['primary_key'],
                        'order': item['order'],
                        'item': item['data'],
                    }
                )



import collections
import logging
import pickle
import time

from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async
from channels.layers import get_channel_layer
from django.core import exceptions as django_exceptions
from django.db import IntegrityError, connection, transaction
from django.http import Http404
from django.utils import timezone
from rest_framework import request as api_request

from . import exceptions, models
from .connection import get_queryobserver_settings
from .protocol import *

# Logger.
logger = logging.getLogger(__name__)

# Observable method options attribute name prefix.
OBSERVABLE_OPTIONS_PREFIX = 'observable_'
# Maximum number of retries in case of concurrent observer creates.
MAX_INTEGRITY_ERROR_RETRIES = 3


class Options:
    """Query observer options."""

    # Valid change detection types.
    CHANGE_DETECTION_PUSH = 'push'
    CHANGE_DETECTION_POLL = 'poll'

    def __init__(self, viewset, viewset_method):
        self._viewset = viewset
        self._viewset_method = viewset_method

        # Determine the primary key.
        self.primary_key = self.get_option('primary_key')
        if self.primary_key is None:
            # Primary key attribute is not defined, attempt to autodiscover it from the queryset.
            try:
                self.primary_key = viewset.get_queryset().model._meta.pk.name
            except AssertionError:
                # No queryset is defined.
                raise exceptions.MissingPrimaryKey(
                    "Observable method does not define a primary key and the viewset "
                    "does not provide a queryset. Define a queryset or use the primary_key "
                    "decorator."
                )

        # Determine change detection type.
        self.change_detection = self.get_option(
            'change_detection', Options.CHANGE_DETECTION_PUSH
        )
        self.poll_interval = self.get_option('poll_interval')
        self.dependencies = self.get_option('dependencies')

    def get_option(self, name, default=None):
        return getattr(
            self._viewset_method,
            '{}{}'.format(OBSERVABLE_OPTIONS_PREFIX, name),
            default,
        )


class QueryObserver:
    """Query observer.

    A query observer observes a specific viewset for changes and propagates these
    changes to all interested subscribers.
    """

    def __init__(self, request):
        """Create new query observer.

        :param request: A `queryobserver.request.Request` instance
        """

        # Obtain a serializer by asking the viewset to provide one. We instantiate the
        # viewset with a fake request, so that the viewset methods work as expected.
        viewset = request.viewset_class()
        viewset.request = api_request.Request(request)
        viewset.request.method = request.method
        viewset.format_kwarg = None
        viewset.args = request.args
        viewset.kwargs = request.kwargs
        self._request = request
        self._viewset = viewset
        self._viewset_method = getattr(viewset, request.viewset_method)
        self._meta = Options(viewset, self._viewset_method)

    @property
    def id(self):
        """Unique observer identifier."""
        return self._request.observe_id

    def _get_logging_extra(self, duration=None, results=None):
        """Extra information for logger."""
        return {
            'duration': duration,
            'results': results,
            'observer_id': self.id,
            'viewset': '{}.{}'.format(
                self._request.viewset_class.__module__,
                self._request.viewset_class.__name__,
            ),
            'method': self._request.viewset_method,
            'path': self._request.path,
            'get': self._request.GET,
        }

    def _get_logging_id(self):
        """Get logging identifier."""
        return "{}.{}/{}".format(
            self._request.viewset_class.__module__,
            self._request.viewset_class.__name__,
            self._request.viewset_method,
        )

    def _warning(self, msg, duration=None, results=None):
        """Log warnings."""
        logger.warning(
            "{} ({})".format(msg, self._get_logging_id()),
            extra=self._get_logging_extra(duration=duration, results=results),
        )

    def subscribe(self, session_id):
        """Initialize observer and register subscriber.

        :param session_id: Subscriber's session identifier
        """
        try:
            change_detection = self._meta.change_detection
            if change_detection not in [
                Options.CHANGE_DETECTION_PUSH,
                Options.CHANGE_DETECTION_POLL,
            ]:
                raise NotImplementedError(
                    "Change detection mechanism '{}' not implemented.".format(
                        change_detection
                    )
                )

            poll_interval = (
                self._meta.poll_interval
                if change_detection == Options.CHANGE_DETECTION_POLL
                else None
            )

            dependencies = self._meta.dependencies

            viewset_results = self._viewset_results()

            # Subscribe to observer in a single query. First, create an
            # observer, then create a subscriber, and finally subscribe to
            # the observer. If already subscribed, ignore the conflict.
            for retry in range(MAX_INTEGRITY_ERROR_RETRIES):
                is_subscribed = False
                cursor = connection.cursor()
                try:
                    cursor.execute(
                        """
                        WITH inserted_observer AS (
                            INSERT into {observer_table} ("id", "request", "poll_interval")
                            VALUES (%(observer_id)s, %(request)s, %(poll_interval)s)
                            ON CONFLICT DO NOTHING
                        ), inserted_subscriber AS (
                            INSERT into {subscriber_table} ("session_id", "created")
                            VALUES (%(subscriber_id)s, NOW())
                            ON CONFLICT DO NOTHING
                        )
                        INSERT INTO {observer_subscribers_table} ("observer_id", "subscriber_id")
                        VALUES (%(observer_id)s, %(subscriber_id)s)
                        """.format(
                            observer_table=models.Observer._meta.db_table,
                            subscriber_table=models.Subscriber._meta.db_table,
                            observer_subscribers_table=models.Observer.subscribers.through._meta.db_table,
                        ),
                        params={
                            'observer_id': self.id,
                            'request': pickle.dumps(self._request),
                            'poll_interval': poll_interval,
                            'subscriber_id': session_id,
                        },
                    )
                    is_subscribed = True
                except IntegrityError as err:
                    msg = str(err)
                    if (
                        'Key (observer_id, subscriber_id)' in msg
                        and 'already exists' in msg
                    ):
                        # Subscriber already subscribed, we're good.
                        is_subscribed = True
                    elif (
                        'Key (observer_id)' in msg or 'Key (subscriber_id)' in msg
                    ) and 'not present in table' in msg:
                        # Could not subscribe because observer, subscriber or
                        # both are missing, retry.
                        if retry == MAX_INTEGRITY_ERROR_RETRIES - 1:
                            raise
                    else:
                        raise
                finally:
                    cursor.close()

                if is_subscribed:
                    break

            # Determine who should notify us based on the configured change
            # detection mechanism.
            if change_detection == Options.CHANGE_DETECTION_PUSH:
                if dependencies:
                    tables = [model._meta.db_table for model in dependencies]
                else:
                    tables = [self._viewset.get_queryset().model._meta.db_table]

                # Register table dependencies for push observables.
                for table in tables:
                    try:
                        models.Dependency.objects.get_or_create(
                            observer_id=self.id, table=table
                        )
                    except models.Observer.DoesNotExist:
                        # The observer was removed before dependency tables
                        # were created.
                        return viewset_results

            elif self._meta.change_detection == Options.CHANGE_DETECTION_POLL:
                # Register poller.
                async_to_sync(get_channel_layer().send)(
                    CHANNEL_MAIN,
                    {
                        'type': TYPE_POLL,
                        'observer': self.id,
                        'interval': self._meta.poll_interval,
                    },
                )

            self._evaluate(viewset_results)

        except Exception:
            logger.exception(
                "Error while subscribing to observer ({})".format(
                    self._get_logging_id()
                ),
                extra=self._get_logging_extra(),
            )

        return viewset_results

    async def evaluate(self):
        """Evaluate the query observer.

        :param return_emitted: True if the emitted diffs should be returned (testing only)
        """

        @database_sync_to_async
        def remove_subscribers():
            models.Observer.subscribers.through.objects.filter(
                observer_id=self.id
            ).delete()

        @database_sync_to_async
        def get_subscriber_sessions():
            return list(
                models.Observer.subscribers.through.objects.filter(observer_id=self.id)
                .distinct('subscriber_id')
                .values_list('subscriber_id', flat=True)
            )

        try:
            settings = get_queryobserver_settings()

            start = time.time()
            # Evaluate the observer
            added, changed, removed = await database_sync_to_async(self._evaluate)()
            duration = time.time() - start

            # Log slow observers.
            if duration > settings['warnings']['max_processing_time']:
                self._warning("Slow observed viewset", duration=duration)

            # Remove subscribers of really slow observers.
            if duration > settings['errors']['max_processing_time']:
                logger.error(
                    "Removing subscribers to extremely slow observed viewset ({})".format(
                        self._get_logging_id()
                    ),
                    extra=self._get_logging_extra(duration=duration),
                )
                await remove_subscribers()

            if self._meta.change_detection == Options.CHANGE_DETECTION_POLL:
                # Register poller.
                await get_channel_layer().send(
                    CHANNEL_MAIN,
                    {
                        'type': TYPE_POLL,
                        'observer': self.id,
                        'interval': self._meta.poll_interval,
                    },
                )

            message = {
                'type': TYPE_ITEM_UPDATE,
                'observer': self.id,
                'primary_key': self._meta.primary_key,
                'added': added,
                'changed': changed,
                'removed': removed,
            }

            # Only generate notifications in case there were any changes.
            if added or changed or removed:
                for session_id in await get_subscriber_sessions():
                    await get_channel_layer().group_send(
                        GROUP_SESSIONS.format(session_id=session_id), message
                    )

        except Exception:
            logger.exception(
                "Error while evaluating observer ({})".format(self._get_logging_id()),
                extra=self._get_logging_extra(),
            )

    def _viewset_results(self):
        """Parse results from the viewset response."""
        results = []
        try:
            response = self._viewset_method(
                self._viewset.request, *self._request.args, **self._request.kwargs
            )

            if response.status_code == 200:
                results = response.data

                if not isinstance(results, list):
                    if isinstance(results, dict):
                        # XXX: This can incidently match if a single
                        # object has results key
                        if 'results' in results and isinstance(
                            results['results'], list
                        ):
                            # Support paginated results.
                            results = results['results']
                        else:
                            results.setdefault(self._meta.primary_key, 1)
                            results = [collections.OrderedDict(results)]
                    else:
                        raise ValueError(
                            "Observable views must return a dictionary or a list of dictionaries!"
                        )
        except Http404:
            pass
        except django_exceptions.ObjectDoesNotExist:
            # The evaluation may fail when certain dependent objects (like users) are removed
            # from the database. In this case, the observer is stopped.
            pass

        return results

    def _evaluate(self, viewset_results=None):
        """Evaluate query observer.

        :param viewset_results: Objects returned by the viewset query
        """
        if viewset_results is None:
            viewset_results = self._viewset_results()

        try:
            observer = models.Observer.objects.get(id=self.id)
            # Do not evaluate the observer if there are no subscribers
            if observer.subscribers.count() == 0:
                return (None, None, None)

            # Update last evaluation time.
            models.Observer.objects.filter(id=self.id).update(
                last_evaluation=timezone.now()
            )

            # Log viewsets with too much output.
            max_result = get_queryobserver_settings()['warnings']['max_result_length']
            if len(viewset_results) > max_result:
                self._warning(
                    "Observed viewset returns too many results",
                    results=len(viewset_results),
                )
            new_results = collections.OrderedDict()
            for order, item in enumerate(viewset_results):
                if not isinstance(item, dict):
                    raise ValueError(
                        "Observable views must return a dictionary or a list of dictionaries!"
                    )

                item = {'order': order, 'data': item}

                try:
                    new_results[str(item['data'][self._meta.primary_key])] = item
                except KeyError:
                    raise KeyError(
                        "Observable view did not return primary key field '{}'!".format(
                            self._meta.primary_key
                        )
                    )

            # Process difference between old results and new results.
            added, changed = [], []
            new_ids = list(new_results.keys())

            removed_qs = observer.items.exclude(primary_key__in=new_results.keys())
            removed = list(removed_qs.values('order', 'data'))

            maybe_changed_qs = observer.items.filter(primary_key__in=new_results.keys())

            with transaction.atomic():
                # Removed items.
                removed_qs.delete()

                # Defer unique ordering constraint before processing order updates.
                # NOTE: The name of the constrait is generated by Django ORM.
                with connection.cursor() as cursor:
                    cursor.execute(
                        "SET CONSTRAINTS rest_framework_reactive_item_observer_id_order_9b8adde6_uniq DEFERRED"
                    )

                # Changed items.
                for item_id, old_order, old_data in maybe_changed_qs.values_list(
                    'primary_key', 'order', 'data'
                ):
                    new_item = new_results[item_id]
                    new_ids.remove(item_id)

                    if new_item['data'] != old_data:
                        changed.append(new_item)
                        observer.items.filter(primary_key=item_id).update(
                            data=new_item['data'], order=new_item['order']
                        )
                    elif new_item['order'] != old_order:
                        # TODO: If only order has changed, don't transmit
                        # full data (needs frontend support).
                        changed.append(new_item)
                        observer.items.filter(primary_key=item_id).update(
                            order=new_item['order']
                        )

                # Added items.
                for item_id in new_ids:
                    item = new_results[item_id]
                    added.append(item)
                    observer.items.create(
                        primary_key=item_id, order=item['order'], data=item['data']
                    )

            return (added, changed, removed)

        except models.Observer.DoesNotExist:
            # Observer removed, ignore evaluation
            return (None, None, None)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return '<QueryObserver: id={id} request={request}>'.format(
            id=self.id, request=repr(self._request)
        )


def remove_subscriber(session_id, observer_id):
    """Remove subscriber from the given observer.

    :param session_id: Subscriber's session identifier
    :param observer_id: Observer identifier
    """
    models.Observer.subscribers.through.objects.filter(
        subscriber_id=session_id, observer_id=observer_id
    ).delete()



import functools
import inspect

from rest_framework import response

from . import observer
from . import request as observer_request


def observable(
    _method_or_viewset=None, poll_interval=None, primary_key=None, dependencies=None
):
    """Make ViewSet or ViewSet method observable.

    Decorating a ViewSet class is the same as decorating its `list` method.

    If decorated method returns a response containing a list of items, it must
    use the provided `LimitOffsetPagination` for any pagination. In case a
    non-list response is returned, the resulting item will be wrapped into a
    list.

    When multiple decorators are used, `observable` must be the first one to be
    applied as it needs access to the method name.

    :param poll_interval: Configure given observable as a polling observable
    :param primary_key: Primary key for tracking observable items
    :param dependencies: List of ORM to register as dependencies for
        orm_notify. If None the observer will subscribe to notifications from
        the queryset model.

    """
    if poll_interval and dependencies:
        raise ValueError('Only one of poll_interval and dependencies arguments allowed')

    def decorator_observable(method_or_viewset):

        if inspect.isclass(method_or_viewset):
            list_method = getattr(method_or_viewset, 'list', None)
            if list_method is not None:
                method_or_viewset.list = observable(
                    list_method,
                    poll_interval=poll_interval,
                    primary_key=primary_key,
                    dependencies=dependencies,
                )

            return method_or_viewset

        # Do not decorate an already observable method twice.
        if getattr(method_or_viewset, 'is_observable', False):
            return method_or_viewset

        @functools.wraps(method_or_viewset)
        def wrapper(self, request, *args, **kwargs):
            if observer_request.OBSERVABLE_QUERY_PARAMETER in request.query_params:
                # TODO: Validate the session identifier.
                session_id = request.query_params[
                    observer_request.OBSERVABLE_QUERY_PARAMETER
                ]

                # Create request and subscribe the session to given observer.
                request = observer_request.Request(
                    self.__class__, method_or_viewset.__name__, request, args, kwargs
                )

                # Initialize observer and subscribe.
                instance = observer.QueryObserver(request)
                data = instance.subscribe(session_id)

                return response.Response({'observer': instance.id, 'items': data})
            else:
                # Non-reactive API.
                return method_or_viewset(self, request, *args, **kwargs)

        wrapper.is_observable = True

        if poll_interval is not None:
            wrapper.observable_change_detection = observer.Options.CHANGE_DETECTION_POLL
            wrapper.observable_poll_interval = poll_interval

        if primary_key is not None:
            wrapper.observable_primary_key = primary_key

        if dependencies is not None:
            wrapper.observable_dependencies = dependencies

        return wrapper

    if _method_or_viewset is None:
        return decorator_observable
    else:
        return decorator_observable(_method_or_viewset)



"""Application configuration."""
from django.apps import AppConfig


class BaseConfig(AppConfig):
    """Application configuration."""

    name = 'rest_framework_reactive'

    def ready(self):
        """Perform application initialization."""
        # Connect all signals.
        from . import signals



# Channel used for lightweight controller tasks.
CHANNEL_MAIN = 'rest_framework_reactive.main'
# Channel used for CPU-bound observers evaluation.
CHANNEL_WORKER = 'rest_framework_reactive.worker'
# Group used for individual sessions.
GROUP_SESSIONS = 'rest_framework_reactive.session.{session_id}'

# Message type for observer evaluation.
TYPE_EVALUATE = 'observer.evaluate'
# Message type for ORM table change notifications.
TYPE_ORM_NOTIFY = 'observer.orm_notify'
# Message type for polling observable evaluation.
TYPE_POLL = 'observer.poll'
# Message type for observer item updates.
TYPE_ITEM_UPDATE = 'observer.update'

ORM_NOTIFY_KIND_CREATE = 'create'
ORM_NOTIFY_KIND_UPDATE = 'update'
ORM_NOTIFY_KIND_DELETE = 'delete'



from rest_framework import response, views

from . import observer


class QueryObserverUnsubscribeView(views.APIView):
    def post(self, request):
        """Handle a query observer unsubscription request."""
        try:
            observer_id = request.query_params['observer']
            session_id = request.query_params['subscriber']
        except KeyError:
            return response.Response(status=400)

        observer.remove_subscriber(session_id, observer_id)
        return response.Response()



from channels.routing import ChannelNameRouter, ProtocolTypeRouter, URLRouter
from django.urls import path

from .consumers import ClientConsumer, MainConsumer, WorkerConsumer
from .protocol import CHANNEL_MAIN, CHANNEL_WORKER

application = ProtocolTypeRouter(
    {
        # Client-facing consumers.
        'websocket': URLRouter(
            [
                # To change the prefix, you can import ClientConsumer in your custom
                # Channels routing definitions instead of using these defaults.
                path('ws/<slug:subscriber_id>', ClientConsumer)
            ]
        ),
        # Background worker consumers.
        'channel': ChannelNameRouter(
            {CHANNEL_MAIN: MainConsumer, CHANNEL_WORKER: WorkerConsumer}
        ),
    }
)



"""
The serializers defined here are only used during testing.
"""
from rest_framework import serializers

from . import models


class ExampleItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.ExampleItem
        fields = ('id', 'enabled', 'name')


class ExampleSubItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.ExampleSubItem
        fields = ('id', 'parent', 'enabled')



"""
Django settings for running tests for django-rest-framework-reactive package.

"""
import os

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))

SECRET_KEY = 'secret'

DEBUG = True

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

# Apps from this project
PROJECT_APPS = (
    'rest_framework_reactive',
    'drfr_test_app.apps.QueryObserverTestsConfig',
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'rest_framework',
    'channels',
    'guardian',
) + PROJECT_APPS

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
            ]
        },
    }
]

AUTHENTICATION_BACKENDS = ('django.contrib.auth.backends.ModelBackend',)

ANONYMOUS_USER_NAME = 'public'

# Get the current Tox testing environment
# NOTE: This is useful for concurrently running tests with separate environments
toxenv = os.environ.get('TOXENV', '')

# Check if PostgreSQL settings are set via environment variables
pgname = os.environ.get('DRFR_POSTGRESQL_NAME', 'drfr')
pguser = os.environ.get('DRFR_POSTGRESQL_USER', 'drfr')
pghost = os.environ.get('DRFR_POSTGRESQL_HOST', 'localhost')
pgport = int(os.environ.get('DRFR_POSTGRESQL_PORT', 55435))

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': pgname,
        'USER': pguser,
        'HOST': pghost,
        'PORT': pgport,
        'TEST': {'NAME': 'drfr_test' + toxenv},
    }
}

STATIC_URL = '/static/'

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_FILTER_BACKENDS': (
        'django_filters.rest_framework.backends.DjangoFilterBackend',
        'rest_framework.filters.OrderingFilter',
    ),
}

CHANNEL_LAYERS = {'default': {'BACKEND': 'channels.layers.InMemoryChannelLayer'}}

ASGI_APPLICATION = 'rest_framework_reactive.routing.application'

DJANGO_REST_FRAMEWORK_REACTIVE = {
    # Set throttle rate to zero during tests as otherwise they can be delayed
    # and cause test timeouts.
    'throttle_rate': 0
}



from django.core.cache import cache
from django.core.management.base import BaseCommand

from ... import models
from ...consumers import THROTTLE_CACHE_PREFIX


class Command(BaseCommand):
    """Clear observer state."""

    help = "Clear observer state: delete all observers and subscribers."

    def handle(self, *args, **options):
        """Command handle."""
        models.Observer.objects.all().delete()
        models.Subscriber.objects.all().delete()

        for cache_key in cache.keys(search='{}*'.format(THROTTLE_CACHE_PREFIX)):
            cache.delete(cache_key)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2019-02-15 09:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [('rest_framework_reactive', '0002_defer_order_constraint')]

    operations = [
        migrations.RemoveField(model_name='observer', name='status'),
        migrations.AlterField(
            model_name='observer',
            name='last_evaluation',
            field=models.DateTimeField(null=True),
        ),
    ]



import asyncio

import async_timeout
import pytest
from channels.db import database_sync_to_async
from channels.layers import get_channel_layer
from channels.routing import URLRouter
from channels.testing import ApplicationCommunicator, WebsocketCommunicator
from django.urls import path
from django.core.cache import cache
from django.test import override_settings
from rest_framework import request as api_request
from rest_framework import test as api_test

from drfr_test_app import models, views
from rest_framework_reactive import models as observer_models
from rest_framework_reactive import request as observer_request
from rest_framework_reactive.consumers import (
    ClientConsumer,
    MainConsumer,
    WorkerConsumer,
    throttle_cache_key,
)
from rest_framework_reactive.observer import QueryObserver
from rest_framework_reactive.protocol import *

# Create test request factory.
factory = api_test.APIRequestFactory()


def create_request(viewset_class, **kwargs):
    request = observer_request.Request(
        viewset_class, 'list', api_request.Request(factory.get('/', kwargs))
    )

    return request


@database_sync_to_async
def assert_subscribers(num, observer_id=None):
    """Test the number of subscribers."""
    if observer_id:
        observer = observer_models.Observer.objects.get(id=observer_id)
        assert observer.subscribers.all().count() == num
    else:
        assert observer_models.Subscriber.objects.all().count() == num


@pytest.mark.django_db(transaction=True)
@pytest.mark.asyncio
async def test_worker_and_client():
    client_consumer = URLRouter([path('ws/<slug:subscriber_id>', ClientConsumer)])

    client = WebsocketCommunicator(client_consumer, '/ws/test-session')
    main = ApplicationCommunicator(
        MainConsumer, {'type': 'channel', 'channel': CHANNEL_MAIN}
    )
    worker = ApplicationCommunicator(
        WorkerConsumer, {'type': 'channel', 'channel': CHANNEL_WORKER}
    )

    # Connect client.
    status, _ = await client.connect()
    assert status is True

    # Create an observer.
    @database_sync_to_async
    def create_observer():
        observer = QueryObserver(
            create_request(views.PaginatedViewSet, offset=0, limit=10)
        )
        items = observer.subscribe('test-session')
        assert not items
        return observer.id

    observer_id = await create_observer()
    await assert_subscribers(1)
    await assert_subscribers(1, observer_id)

    # Create a single model instance for the observer model.
    @database_sync_to_async
    def create_model():
        return models.ExampleItem.objects.create(enabled=True, name="hello world").pk

    primary_key = await create_model()
    channel_layer = get_channel_layer()

    async with async_timeout.timeout(1):
        # Check that ORM signal was generated.
        notify = await channel_layer.receive(CHANNEL_MAIN)
        assert notify['type'] == TYPE_ORM_NOTIFY
        assert notify['kind'] == ORM_NOTIFY_KIND_CREATE
        assert notify['primary_key'] == str(primary_key)

        # Propagate notification to worker.
        await main.send_input(notify)

        # Check that observer evaluation was requested.
        notify = await channel_layer.receive(CHANNEL_WORKER)
        assert notify['type'] == TYPE_EVALUATE
        assert notify['observer'] == observer_id

        # Propagate notification to worker.
        await worker.send_input(notify)
        response = await client.receive_json_from()
        assert response['msg'] == 'added'
        assert response['primary_key'] == 'id'
        assert response['order'] == 0
        assert response['item'] == {'id': 1, 'enabled': True, 'name': 'hello world'}

    # No other messages should be sent.
    assert await client.receive_nothing() is True
    await client.disconnect()

    # Ensure that subscriber has been removed.
    await assert_subscribers(0)

    async with async_timeout.timeout(1):
        # Run observer again and it should skip evaluation because there are no more subscribers.
        await worker.send_input({'type': TYPE_EVALUATE, 'observer': observer_id})

    assert await worker.receive_nothing() is True


@pytest.mark.asyncio
async def test_poll_observer():
    main = ApplicationCommunicator(
        MainConsumer, {'type': 'channel', 'channel': CHANNEL_MAIN}
    )
    await main.send_input({'type': TYPE_POLL, 'observer': 'test', 'interval': 2})
    channel_layer = get_channel_layer()

    # Nothing should be received in the first second.
    with pytest.raises(asyncio.TimeoutError):
        async with async_timeout.timeout(1):
            await channel_layer.receive(CHANNEL_WORKER)
            assert False

    # Then after another second we should get a notification.
    async with async_timeout.timeout(2):
        notify = await channel_layer.receive(CHANNEL_WORKER)
        assert notify['type'] == TYPE_EVALUATE
        assert notify['observer'] == 'test'


@pytest.mark.django_db(transaction=True)
@pytest.mark.asyncio
async def test_poll_observer_integration():
    client_consumer = URLRouter([path('ws/<slug:subscriber_id>', ClientConsumer)])

    client = WebsocketCommunicator(client_consumer, '/ws/test-session')
    main = ApplicationCommunicator(
        MainConsumer, {'type': 'channel', 'channel': CHANNEL_MAIN}
    )
    worker = ApplicationCommunicator(
        WorkerConsumer, {'type': 'channel', 'channel': CHANNEL_WORKER}
    )

    # Connect client.
    status, _ = await client.connect()
    assert status is True

    # Create an observer.
    @database_sync_to_async
    def create_observer():
        observer = QueryObserver(create_request(views.PollingObservableViewSet))
        items = observer.subscribe('test-session')
        assert len(items) == 1
        return observer.id

    observer_id = await create_observer()
    await assert_subscribers(1)
    await assert_subscribers(1, observer_id)

    channel_layer = get_channel_layer()

    async with async_timeout.timeout(1):
        # Ensure that a notification message was sent to the poller.
        notify = await channel_layer.receive(CHANNEL_MAIN)
        assert notify['type'] == TYPE_POLL
        assert notify['interval'] == 2
        assert notify['observer'] == observer_id

        # Dispatch notification to poller as our poller uses a dummy queue.
        await main.send_input(notify)

    # Nothing should be received in the first second.
    with pytest.raises(asyncio.TimeoutError):
        async with async_timeout.timeout(1):
            await channel_layer.receive(CHANNEL_WORKER)
            assert False

    async with async_timeout.timeout(2):
        # Then after another second we should get a notification.
        notify = await channel_layer.receive(CHANNEL_WORKER)

        # Dispatch notification to worker.
        await worker.send_input(notify)

        # Ensure another notification message was sent to the poller.
        notify = await channel_layer.receive(CHANNEL_MAIN)
        assert notify['type'] == TYPE_POLL
        assert notify['interval'] == 2
        assert notify['observer'] == observer_id

        # Ensure client got notified of changes.
        response = await client.receive_json_from()
        assert response['msg'] == 'changed'
        assert response['primary_key'] == 'id'
        assert response['order'] == 0
        assert response['item']['static'].startswith('This is a polling observable:')

    # No other messages should be sent.
    assert await client.receive_nothing() is True
    await client.disconnect()


@pytest.mark.django_db(transaction=True)
@pytest.mark.asyncio
async def test_throttle_observer():
    with override_settings(DJANGO_REST_FRAMEWORK_REACTIVE={'throttle_rate': 2}):
        channel_layer = get_channel_layer()
        main = ApplicationCommunicator(
            MainConsumer, {'type': 'channel', 'channel': CHANNEL_MAIN}
        )
        worker = ApplicationCommunicator(
            WorkerConsumer, {'type': 'channel', 'channel': CHANNEL_WORKER}
        )

        @database_sync_to_async
        def create_observer():
            observer = QueryObserver(
                create_request(views.ExampleItemViewSet, offset=0, limit=10)
            )
            items = observer.subscribe('test-session')
            return observer.id

        @database_sync_to_async
        def get_last_evaluation(observer_id):
            return observer_models.Observer.objects.get(id=observer_id).last_evaluation

        def throttle_count(observer_id, value=None):
            cache_key = throttle_cache_key(observer_id)
            if value is None:
                return cache.get(cache_key)
            else:
                return cache.set(cache_key, value)

        observer_id = await create_observer()

        # Test that observer is evaluated
        async with async_timeout.timeout(1):
            await worker.send_input({'type': TYPE_EVALUATE, 'observer': observer_id})

            # Nothing should be in the main worker
            assert await main.receive_nothing()

            # Get last evaluation time for later comparisson
            last_evaluation = await get_last_evaluation(observer_id)

            # Throttle count should be 1 (one process)
            assert throttle_count(observer_id) == 1

            # Ensure last_evaluated time change before the next test
            await asyncio.sleep(0.001)

            # Observer evaluation is delayed while another is evaluated
            await worker.send_input({'type': TYPE_EVALUATE, 'observer': observer_id})

            # Delayed observer should be scheduled
            notify = await channel_layer.receive(CHANNEL_MAIN)
            assert notify['type'] == TYPE_POLL
            assert notify['observer'] == observer_id

            # Throttle count should be 2 (two processes)
            assert throttle_count(observer_id) == 2

            # Observer should not be evaluated
            assert last_evaluation == await get_last_evaluation(observer_id)

            # Observer evaluation is discarded when delayed observer scheduled
            await worker.send_input({'type': TYPE_EVALUATE, 'observer': observer_id})

            # Nothing should be in the main worker
            assert await main.receive_nothing()

            # Observer should not be evaluated
            assert last_evaluation == await get_last_evaluation(observer_id)

            # Throttle count should be 3 (three processes)
            assert throttle_count(observer_id) == 3



import pickle

from django import test
from django.contrib.auth import models as auth_models
from guardian import shortcuts
from rest_framework import request as api_request
from rest_framework import test as api_test

from drfr_test_app import models, views
from rest_framework_reactive import models as observer_models
from rest_framework_reactive import request as observer_request
from rest_framework_reactive.observer import QueryObserver, remove_subscriber

# Create test request factory.
factory = api_test.APIRequestFactory()


def request(viewset_class, **kwargs):
    request = observer_request.Request(
        viewset_class, 'list', api_request.Request(factory.get('/', kwargs))
    )

    # Simulate serialization.
    return pickle.loads(pickle.dumps(request))


class QueryObserversTestCase(test.TestCase):
    def test_paginated_viewset(self):
        observer = QueryObserver(request(views.PaginatedViewSet, offset=0, limit=10))
        items = observer.subscribe('test-session')

        self.assertEqual(len(items), 0)

        items = []
        for index in range(20):
            items.append(
                models.ExampleItem.objects.create(name='Example', enabled=True)
            )

        # Evaluate the observer again (in reality this would be done automatically, triggered by signals
        # from Django ORM).
        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 10)
        expected_serialized_item = {
            'id': items[0].pk,
            'name': items[0].name,
            'enabled': items[0].enabled,
        }
        self.assertEqual(added[0]['data'], expected_serialized_item)
        self.assertEqual(len(changed), 0)
        self.assertEqual(len(removed), 0)

    def test_item_serialization(self):
        item = models.ExampleItem.objects.create(name='Example', enabled=True)
        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item,
        )

        observer = QueryObserver(request(views.ExampleItemViewSet))
        items = observer.subscribe('test-session')

        # Ensure items can be serialized into JSON.
        # json.dumps(items)

    def test_observe_viewset(self):
        # Create a request and an observer for it.
        observer = QueryObserver(request(views.ExampleItemViewSet))
        items = observer.subscribe('test-session')

        self.assertEqual(
            observer.id,
            'fa87c86f1e032942b699e9902ac38ca232ce3566724b3891914c80083b676ed4',
        )
        self.assertEqual(len(items), 0)

        # Add an item into the database.
        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = True
        item.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item,
        )

        # Evaluate the observer again (in reality this would be done automatically, triggered by signals
        # from Django ORM).
        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 1)
        expected_serialized_item = {
            'id': item.pk,
            'name': item.name,
            'enabled': item.enabled,
        }
        self.assertEqual(added[0]['data'], expected_serialized_item)
        self.assertEqual(len(changed), 0)
        self.assertEqual(len(removed), 0)

        # Change the existing item.
        item.enabled = False
        expected_serialized_item['enabled'] = False
        item.save()

        added, changed, removed = observer._evaluate()
        self.assertEqual(len(added), 0)
        self.assertEqual(len(changed), 1)
        self.assertEqual(changed[0]['data'], expected_serialized_item)
        self.assertEqual(len(removed), 0)

        # Remove the first item.
        item.delete()

        # Add another two items.
        item2 = models.ExampleItem()
        item2.name = 'Example 2'
        item2.enabled = True
        item2.save()

        item3 = models.ExampleItem()
        item3.name = 'Example 3'
        item3.enabled = True
        item3.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item2,
        )
        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item3,
        )

        added, changed, removed = observer._evaluate()
        self.assertEqual(len(added), 2)
        self.assertEqual(
            added[0]['data'],
            {'id': item2.pk, 'name': item2.name, 'enabled': item2.enabled},
        )
        self.assertEqual(
            added[1]['data'],
            {'id': item3.pk, 'name': item3.name, 'enabled': item3.enabled},
        )
        self.assertEqual(len(changed), 0)
        self.assertEqual(len(removed), 1)
        self.assertEqual(removed[0]['data'], expected_serialized_item)

    def test_conditions(self):
        observer = QueryObserver(request(views.ExampleItemViewSet, enabled=True))
        items = observer.subscribe('test-session')

        self.assertEqual(
            observer.id,
            '5333b85599fd24ed4e2f7eeaefb599cbbd39894b437e9b9d3b80d5d21639b4bb',
        )
        self.assertEqual(len(items), 0)

        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = False
        item.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item,
        )

        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 0)
        self.assertEqual(len(changed), 0)
        self.assertEqual(len(removed), 0)

        item.enabled = True
        item.save()

        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 1)
        self.assertEqual(
            added[0]['data'],
            {'id': item.pk, 'name': item.name, 'enabled': item.enabled},
        )
        self.assertEqual(len(changed), 0)
        self.assertEqual(len(removed), 0)

    def test_joins(self):
        # Create some items so that we get a valid query (otherwise the query would be empty as django-guardian
        # would discover that the user doesn't have permissions to get any items).
        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = False
        item.save()

        subitem = models.ExampleSubItem(parent=item, enabled=True)
        subitem.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item,
        )
        shortcuts.assign_perm(
            'rest_framework_reactive.view_examplesubitem',
            auth_models.AnonymousUser(),
            subitem,
        )

        observer = QueryObserver(
            request(views.ExampleSubItemViewSet, parent__enabled=True)
        )
        items = observer.subscribe('test-session')

        self.assertEqual(
            observer.id,
            '92b1698976bf1e04d155f9c60ac74c054ef872f547a59d771fc3c046998bbba8',
        )
        self.assertEqual(len(items), 0)

        observer_state = observer_models.Observer.objects.get(pk=observer.id)
        dependencies = observer_state.dependencies.all().values_list('table', flat=True)
        self.assertIn('drfr_test_app_exampleitem', dependencies)
        self.assertIn('drfr_test_app_examplesubitem', dependencies)

    def test_aggregations(self):
        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = False
        item.save()

        m2m_item = models.ExampleM2MItem()
        m2m_item.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item,
        )

        observer = QueryObserver(
            request(views.AggregationTestViewSet, items=[m2m_item.pk])
        )
        observer.subscribe('test-session')

        # There should be a dependency on the intermediate table.
        observer_state = observer_models.Observer.objects.get(pk=observer.id)
        dependencies = observer_state.dependencies.all().values_list('table', flat=True)
        self.assertIn('drfr_test_app_examplem2mitem_items', dependencies)

    def test_order(self):
        observer = QueryObserver(request(views.ExampleItemViewSet, ordering='name'))
        items = observer.subscribe('test-session')

        self.assertEqual(len(items), 0)

        item = models.ExampleItem()
        item.name = 'D'
        item.enabled = False
        item.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item,
        )

        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 1)
        self.assertEqual(
            added[0]['data'],
            {'id': item.pk, 'name': item.name, 'enabled': item.enabled},
        )
        self.assertEqual(added[0]['order'], 0)
        self.assertEqual(len(changed), 0)
        self.assertEqual(len(removed), 0)

        item2 = models.ExampleItem()
        item2.name = 'A'
        item2.enabled = True
        item2.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item2,
        )

        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 1)
        self.assertEqual(
            added[0]['data'],
            {'id': item2.pk, 'name': item2.name, 'enabled': item2.enabled},
        )
        self.assertEqual(added[0]['order'], 0)
        # Check that the first item has changed, because its order has changed.
        self.assertEqual(len(changed), 1)
        self.assertEqual(
            changed[0]['data'],
            {'id': item.pk, 'name': item.name, 'enabled': item.enabled},
        )
        self.assertEqual(changed[0]['order'], 1)
        self.assertEqual(len(removed), 0)

        item3 = models.ExampleItem()
        item3.name = 'C'
        item3.enabled = True
        item3.save()

        shortcuts.assign_perm(
            'rest_framework_reactive.view_exampleitem',
            auth_models.AnonymousUser(),
            item3,
        )

        added, changed, removed = observer._evaluate()
        self.assertEqual(len(added), 1)
        self.assertEqual(
            added[0]['data'],
            {'id': item3.pk, 'name': item3.name, 'enabled': item3.enabled},
        )
        self.assertEqual(added[0]['order'], 1)
        self.assertEqual(len(changed), 1)
        self.assertEqual(
            changed[0]['data'],
            {'id': item.pk, 'name': item.name, 'enabled': item.enabled},
        )
        self.assertEqual(changed[0]['order'], 2)
        self.assertEqual(len(removed), 0)

        # Check order change between two existing items.

        item.name = 'B'
        item.save()

        added, changed, removed = observer._evaluate()

        self.assertEqual(len(added), 0)
        self.assertEqual(len(changed), 2)
        self.assertEqual(
            changed[0]['data'],
            {'id': item3.pk, 'name': item3.name, 'enabled': item3.enabled},
        )
        self.assertEqual(changed[0]['order'], 2)
        self.assertEqual(
            changed[1]['data'],
            {'id': item.pk, 'name': item.name, 'enabled': item.enabled},
        )
        self.assertEqual(changed[1]['order'], 1)
        self.assertEqual(len(removed), 0)

    def test_no_dependencies(self):
        observer = QueryObserver(request(views.NoDependenciesViewSet))
        items = observer.subscribe('test-session')

        # Observer is created even when there are no dependencies.
        self.assertEqual(len(items), 1)
        self.assertTrue(observer_models.Observer.objects.exists())

    def test_remove_subscriber(self):
        # Simulate opening a WebSocket.
        # This should create an observer and subscribe a subscriber.
        query_observer = QueryObserver(request(views.ExampleItemViewSet))
        items = query_observer.subscribe('test-session')
        observer = observer_models.Observer.objects.get(id=query_observer.id)

        self.assertEqual(observer_models.Observer.objects.count(), 1)
        self.assertEqual(observer.subscribers.count(), 1)

        # Simulate closing a WebSocket.
        # This removes the subscriber but the observer remains.
        remove_subscriber('test-session', observer.id)
        self.assertEqual(observer_models.Observer.objects.count(), 1)
        self.assertEqual(observer.subscribers.count(), 0)

        # Simulate opening a WebSocket again.
        # This should add a subscriber to existing observer.
        query_observer = QueryObserver(request(views.ExampleItemViewSet))
        items = query_observer.subscribe('test-session')
        self.assertEqual(observer_models.Subscriber.objects.count(), 1)
        self.assertEqual(observer.subscribers.count(), 1)

    def test_tables_of_empty_results(self):
        models.ExampleItem.objects.all().delete()
        query_observer = QueryObserver(
            request(views.ExampleItemViewSet, offset=10, limit=0)
        )
        items = query_observer.subscribe('test-session')
        print(items)


class QueryObserversTransactionTestCase(test.TransactionTestCase):
    def test_subscribe(self):
        observer = QueryObserver(request(views.ExampleItemViewSet))
        observer_qs = observer_models.Observer.objects

        # Subscribe new subscriber to new observer.
        observer.subscribe('test-session')
        self.assertEqual(observer_qs.count(), 1)
        self.assertEqual(observer_qs.first().subscribers.count(), 1)

        # Subscribe new subscriber to existing observer.
        observer.subscribe('test-session2')
        self.assertEqual(observer_qs.count(), 1)
        self.assertEqual(observer_qs.first().subscribers.count(), 2)

        # Subscribe existing subscriber to new observer.
        observer_qs.all().delete()
        self.assertEqual(observer_qs.count(), 0)
        self.assertEqual(observer_models.Subscriber.objects.count(), 2)
        observer.subscribe('test-session')
        self.assertEqual(observer_qs.count(), 1)
        self.assertEqual(observer_qs.first().subscribers.count(), 1)

        # Subscribe existing subscriber to existing observer.
        # Note that when a subscriber is already subscribed to the observer,
        # we should ignore duplicate key violation.
        observer.subscribe('test-session')
        self.assertEqual(observer_qs.count(), 1)
        self.assertEqual(observer_qs.first().subscribers.count(), 1)

        # Test that custom SQL query in subscribe serializes the request
        # string the same as Django create.
        observer_obj = observer_qs.first()
        observer_with_same_request = observer_models.Observer.objects.create(
            id='request-comparisson', request=pickle.dumps(observer._request)
        )

        self.assertEqual(
            bytes(observer_obj.request), bytes(observer_with_same_request.request)
        )

    def test_decorate_class(self):
        """Decorating class should pass dependencies to the list method"""
        observer = QueryObserver(request(views.AggregationTestViewSet))
        observer.subscribe('test-session')
        self.assertEqual(
            observer_models.Dependency.objects.filter(observer=observer.id).count(), 2
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-05-02 04:21
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='ExampleItem',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('enabled', models.BooleanField()),
                ('name', models.CharField(max_length=30)),
            ],
            options={
                'ordering': ['pk'],
                'permissions': (('view_exampleitem', 'Can view example item'),),
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='ExampleM2MItem',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('value', models.IntegerField(default=1)),
                (
                    'items',
                    models.ManyToManyField(
                        related_name='items', to='drfr_test_app.ExampleItem'
                    ),
                ),
            ],
            options={'ordering': ['pk']},
        ),
        migrations.CreateModel(
            name='ExampleSubItem',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('enabled', models.BooleanField()),
                (
                    'parent',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='drfr_test_app.ExampleItem',
                    ),
                ),
            ],
            options={
                'ordering': ['pk'],
                'permissions': (('view_examplesubitem', 'Can view example sub item'),),
                'default_permissions': (),
            },
        ),
    ]



import os.path
import setuptools

# Get long description from README.
with open('README.rst', 'r') as fh:
    long_description = fh.read()

# Get package metadata from '__about__.py' file.
about = {}
base_dir = os.path.abspath(os.path.dirname(__file__))
with open(
    os.path.join(base_dir, 'src', 'rest_framework_reactive', '__about__.py'), 'r'
) as fh:
    exec(fh.read(), about)

setuptools.setup(
    name=about['__title__'],
    use_scm_version=True,
    description=about['__summary__'],
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author=about['__author__'],
    author_email=about['__email__'],
    url=about['__url__'],
    license=about['__license__'],
    # Exclude tests from built/installed package.
    packages=setuptools.find_packages(
        'src', exclude=['tests', 'tests.*', '*.tests', '*.tests.*']
    ),
    package_dir={'': 'src'},
    python_requires='>=3.6, <3.9',
    install_requires=['Django~=2.2', 'djangorestframework~=3.9', 'channels~=2.2'],
    extras_require={
        'docs': ['sphinx>=1.3.2', 'sphinx_rtd_theme'],
        'package': ['twine', 'wheel'],
        'test': [
            'django-filter~=2.0',
            'django-guardian>=1.5',
            'django-priority-batch~=2.0',
            'channels-redis~=2.4',
            'pytest~=4.4.1',
            'pytest-django~=3.4.8',
            'pytest-asyncio~=0.10.0',
            'async_timeout~=3.0.1',
            'psycopg2-binary~=2.8',
            'check-manifest',
            'twine',
            'setuptools_scm',
            'black',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='django-rest-framework reactive django',
)



import logging

from asgiref.sync import async_to_sync
from channels.exceptions import ChannelFull
from channels.layers import get_channel_layer
from django import dispatch
from django.db import transaction
from django.db.models import signals as model_signals
from django_priority_batch import PrioritizedBatcher

from .models import Observer
from .protocol import *

# Logger.
logger = logging.getLogger(__name__)

# Global 'in migrations' flag to skip certain operations during migrations.
IN_MIGRATIONS = False


@dispatch.receiver(model_signals.pre_migrate)
def model_pre_migrate(*args, **kwargs):
    """Set 'in migrations' flag."""
    global IN_MIGRATIONS
    IN_MIGRATIONS = True


@dispatch.receiver(model_signals.post_migrate)
def model_post_migrate(*args, **kwargs):
    """Clear 'in migrations' flag."""
    global IN_MIGRATIONS
    IN_MIGRATIONS = False


def notify_observers(table, kind, primary_key=None):
    """Transmit ORM table change notification.

    :param table: Name of the table that has changed
    :param kind: Change type
    :param primary_key: Primary key of the affected instance
    """

    if IN_MIGRATIONS:
        return

    # Don't propagate events when there are no observers to receive them.
    if not Observer.objects.filter(dependencies__table=table).exists():
        return

    def handler():
        """Send a notification to the given channel."""
        try:
            async_to_sync(get_channel_layer().send)(
                CHANNEL_MAIN,
                {
                    'type': TYPE_ORM_NOTIFY,
                    'table': table,
                    'kind': kind,
                    'primary_key': str(primary_key),
                },
            )
        except ChannelFull:
            logger.exception("Unable to notify workers.")

    batcher = PrioritizedBatcher.global_instance()
    if batcher.is_started:
        # If a batch is open, queue the send via the batcher.
        batcher.add(
            'rest_framework_reactive', handler, group_by=(table, kind, primary_key)
        )
    else:
        # If no batch is open, invoke immediately.
        handler()


@dispatch.receiver(model_signals.post_save)
def model_post_save(sender, instance, created=False, **kwargs):
    """Signal emitted after any model is saved via Django ORM.

    :param sender: Model class that was saved
    :param instance: The actual instance that was saved
    :param created: True if a new row was created
    """

    if sender._meta.app_label == 'rest_framework_reactive':
        # Ignore own events.
        return

    def notify():
        table = sender._meta.db_table
        if created:
            notify_observers(table, ORM_NOTIFY_KIND_CREATE, instance.pk)
        else:
            notify_observers(table, ORM_NOTIFY_KIND_UPDATE, instance.pk)

    transaction.on_commit(notify)


@dispatch.receiver(model_signals.post_delete)
def model_post_delete(sender, instance, **kwargs):
    """Signal emitted after any model is deleted via Django ORM.

    :param sender: Model class that was deleted
    :param instance: The actual instance that was removed
    """

    if sender._meta.app_label == 'rest_framework_reactive':
        # Ignore own events.
        return

    def notify():
        table = sender._meta.db_table
        notify_observers(table, ORM_NOTIFY_KIND_DELETE, instance.pk)

    transaction.on_commit(notify)


@dispatch.receiver(model_signals.m2m_changed)
def model_m2m_changed(sender, instance, action, **kwargs):
    """
    Signal emitted after any M2M relation changes via Django ORM.

    :param sender: M2M intermediate model
    :param instance: The actual instance that was saved
    :param action: M2M action
    """

    if sender._meta.app_label == 'rest_framework_reactive':
        # Ignore own events.
        return

    def notify():
        table = sender._meta.db_table
        if action == 'post_add':
            notify_observers(table, ORM_NOTIFY_KIND_CREATE)
        elif action in ('post_remove', 'post_clear'):
            notify_observers(table, ORM_NOTIFY_KIND_DELETE)

    transaction.on_commit(notify)



from django.contrib.postgres.fields import JSONField
from django.db import models


class Observer(models.Model):
    """State of an observer."""

    id = models.CharField(primary_key=True, max_length=64)
    request = models.BinaryField()
    last_evaluation = models.DateTimeField(null=True)
    poll_interval = models.IntegerField(null=True)
    subscribers = models.ManyToManyField('Subscriber')

    def __str__(self):
        return 'id={id}'.format(id=self.id)


class Item(models.Model):
    """Item part of the observer's result set."""

    observer = models.ForeignKey(
        Observer, related_name='items', on_delete=models.CASCADE
    )
    primary_key = models.CharField(max_length=200)
    order = models.IntegerField()
    data = JSONField()

    class Meta:
        unique_together = (('observer', 'order'), ('observer', 'primary_key'))
        ordering = ('observer', 'order')

    def __str__(self):
        return 'primary_key={primary_key} order={order} data={data}'.format(
            primary_key=self.primary_key, order=self.order, data=repr(self.data)
        )


class Dependency(models.Model):
    """Observer's dependency."""

    observer = models.ForeignKey(
        Observer, related_name='dependencies', on_delete=models.CASCADE
    )
    table = models.CharField(max_length=100)

    class Meta:
        unique_together = ('observer', 'table')

    def __str__(self):
        return 'table={table}'.format(table=self.table)


class Subscriber(models.Model):
    """Subscriber to an observer."""

    session_id = models.CharField(primary_key=True, max_length=100)
    created = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return 'session_id={session_id}'.format(session_id=self.session_id)



"""The views defined here are only used during testing."""
import time

from rest_framework import mixins, viewsets
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.response import Response

from rest_framework_reactive.decorators import observable

from . import models, serializers


@observable
class ExampleItemViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet
):

    queryset = models.ExampleItem.objects.all()
    serializer_class = serializers.ExampleItemSerializer
    filterset_fields = ('name', 'enabled', 'items')


@observable(dependencies=[models.ExampleSubItem, models.ExampleItem])
class ExampleSubItemViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet
):

    queryset = models.ExampleSubItem.objects.all()
    serializer_class = serializers.ExampleSubItemSerializer
    filterset_fields = ('parent__enabled', 'enabled')


@observable
class PaginatedViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet
):

    queryset = models.ExampleItem.objects.all()
    serializer_class = serializers.ExampleItemSerializer
    pagination_class = LimitOffsetPagination


@observable(dependencies=[models.ExampleItem, models.ExampleM2MItem.items.through])
class AggregationTestViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet
):

    queryset = models.ExampleItem.objects.all()
    serializer_class = serializers.ExampleItemSerializer
    filterset_fields = ('items',)

    def list(self, request, *args, **kwargs):
        """Filtered query, which just returns a count.

        Such a formulation is used to force the compiler to generate a
        subquery, which uses the M2M relation.
        """
        queryset = self.filter_queryset(self.get_queryset())
        return Response({'count': queryset.count()})


class PollingObservableViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet
):

    queryset = models.ExampleItem.objects.none()
    serializer_class = serializers.ExampleItemSerializer

    @observable(poll_interval=2)
    def list(self, request, *args, **kwargs):
        return Response(
            {'static': 'This is a polling observable: {}'.format(time.time())}
        )


class NoDependenciesViewSet(
    mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet
):

    queryset = models.ExampleItem.objects.none()
    serializer_class = serializers.ExampleItemSerializer

    @observable
    def list(self, request, *args, **kwargs):
        return Response({'static': 'This has no dependencies'})
